# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import shlex
import subprocess
import sys

import nr.fs  # type: ignore

__all__ = ['editor_open', 'edit_text']


def editor_open(filename: str):
  editor = shlex.split(os.getenv('EDITOR', 'vim'))
  return subprocess.call(editor + [filename])


def edit_text(text: str) -> str:
  """
  Opens an editor for the user to modify *text*.
  """

  with nr.fs.tempfile('.yml', dir=os.getcwd(), text=True) as fp:
    fp.write(text)
    fp.close()
    res = editor_open(fp.name)
    if res != 0:
      sys.exit(res)
    with open(fp.name) as src:
      return src.read()
